%{

/* 
   tag.l -- PIKT piktc tag lexer:  deals with machine conditionals and the
            tagging of config file lines; tagging does not apply to
            systems.cfg!

   $Header: /var/cvs/PIKT/pikt/src/piktc/tag.l,v 1.81 2006/11/04 09:59:36 berto Exp $

   This file is part of PIKT ("Problem Informant/Killer Tool").
   Copyright (C) 1998-2006 by Robert Osterlund.

   PIKT is a trademark of the University of Chicago.  All Rights Reserved.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   PIKT is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
   USA.

   Send problem reports, bouquets and brickbats to:  pikt-bugs@pikt.org
 */

#include "../lib/common.h"
#include "piktc.h"
#include "tag.tab.h"

char _taglex_[] = "taglex";

#define UPDATE_INCLUDES		((scanpass == 2) && (inctype == INCPROC) && isupdate_includes && infile && ! incdevnull)

#define UPDATE_INCLUDE		if (UPDATE_INCLUDES) fprintf(inctmpout, "%s", yytext);

#define RETURN(token) taglval.sval = mystrdup(yytext, _taglex_); return token

typedef struct {
	int inctyp;
	char *file;
	char *proc;
	char *tmpfile;
	FILE *tmpout;
	FILE *yyin;
	int fd[2];
	YY_BUFFER_STATE bs;
	pid_t pid;
	int domcr;
	int indnt;
	int lineno;
} incfil;

incfil tagincfiles[MAXIFS+1];

char *tagincfile = (char *) NULL;
char *tagincproc = (char *) NULL;

char *cp;
int ii, jj;

char *execunput;

char *envar, *enval;

void tagerror(const char *);
int tagwrap(void);
void begintagincfile(int, char *, char *);
void pushtagincfile(int, char *, char *, char *, FILE *, FILE *, int *, YY_BUFFER_STATE, pid_t, int, int, int);
void poptagincfile(int *, char **, char **, char **, FILE **, FILE **, int *, YY_BUFFER_STATE *, pid_t *, int *, int *, int *);

%}

W		[ \t]
D		[0-9]
L		[A-Za-z]
DL		[0-9A-Za-z]

%x REM COND INC VRB

%%

<VRB>.*\n?			{ lineno++;
				  UPDATE_INCLUDE;
				  taglval.sval = (char *) mycalloc((size_t)(yyleng+2), sizeof(char), _taglex_);
				  sprintf(taglval.sval, "\t%s", yytext);
				  return _LINE;
				}

<*>^#{W}*(fix|freeze){W}*(\177)?\n {
				  lineno++;
				  domacros = FALSE;
				}
<*>^#{W}*(unfix|unfreeze){W}*(\177)?\n? {
				  lineno++;
				  domacros = TRUE;
				}

<*>^#{W}*indent{W}*(\177)?\n?	{ lineno++;
				  if (indented) {
					logreturn(PIKTLOG, LOG_WARNING, ERRID_INDENT_WITH_OPEN_INDENT, __FILE__, __LINE__, _taglex_, FALSE, "%s, #indent with earlier open #indent at or near line %d: %s", cfgfilecurr, lineno, yytext);
				  }
				  indented = TRUE;
				}
<*>^#{W}*unindent{W}*(\177)?\n?	{ lineno++;
				  if (! indented) {
					logreturn(PIKTLOG, LOG_WARNING, ERRID_UNINDENT_WITHOUT_OPEN_INDENT, __FILE__, __LINE__, _taglex_, FALSE, "%s, #unindent with no earlier open #indent at or near line %d: %s", cfgfilecurr, lineno, yytext);
				  }
				  indented = FALSE;
				}

<*>^#{W}*(include|verbatim){W}+\<[^\>\]\n]+\>{W}+\[[^\n]+\]{W}*(\177)?\n? {
				  lineno++;
				  verbatim = isverbatim(yytext);
				  *strrchr(yytext, ']') = '\0';
				  cp = strchr(yytext, '[') + 1;
				  tagincproc = mystrdup(cp, _taglex_);
				  *strchr(yytext, '>') = '\0';
				  cp = strchr(yytext, '<') + 1;
				  tagincfile = (char *) mymalloc((strlen(cfgdir)+strlen(cp)+2) * sizeof(char), _taglex_);
				  sprintf(tagincfile, "%s/%s", cfgdir, cp);
				  begintagincfile(isupdate_includes ? INCPROC : INCFILE, tagincfile, tagincproc);
				}

<*>^#{W}*(include|verbatim){W}+\"[^\"\]\n]+\"{W}+\[[^\n]+\]{W}*(\177)?\n? {
				  lineno++;
				  verbatim = isverbatim(yytext);
				  *strrchr(yytext, ']') = '\0';
				  cp = strchr(yytext, '[') + 1;
				  tagincproc = mystrdup(cp, _taglex_);
				  *cp = '\0';
				  *strrchr(yytext, '\"') = '\0';
				  cp = strchr(yytext, '\"') + 1;
				  tagincfile = mystrdup(cp, _taglex_);
				  begintagincfile(isupdate_includes ? INCPROC : INCFILE, tagincfile, tagincproc);
				}

<*>^#{W}*(include|verbatim){W}+\[[^\n]+\]{W}*(\177)?\n? {
				  lineno++;
				  verbatim = isverbatim(yytext);
				  *strrchr(yytext, ']') = '\0';
				  cp = strchr(yytext, '[') + 1;
				  tagincproc = mystrdup(cp, _taglex_);
				  tagincfile = (char *) NULL;
				  begintagincfile(INCPROC, tagincfile, tagincproc);
				}

<*>^#{W}*(include|verbatim){W}+\<[^\>\n]+\>{W}*(\177)?\n? {
				  lineno++;
				  verbatim = isverbatim(yytext);
				  tagincproc = (char *) NULL;
				  *strrchr(yytext, '>') = '\0';
				  cp = strchr(yytext, '<') + 1;
				  tagincfile = (char *) mymalloc((strlen(cfgdir)+strlen(cp)+2) * sizeof(char), _taglex_);
				  sprintf(tagincfile, "%s/%s", cfgdir, cp);
				  begintagincfile(INCFILE, tagincfile, tagincproc);
				}

<*>^#{W}*(include|verbatim){W}+\"[^\"\n]+\"{W}*(\177)?\n? {
				  lineno++;
				  verbatim = isverbatim(yytext);
				  tagincproc = (char *) NULL;
				  *strrchr(yytext, '\"') = '\0';
				  cp = strchr(yytext, '\"') + 1;
				  tagincfile = mystrdup(cp, _taglex_);
				  begintagincfile(INCFILE, tagincfile, tagincproc);
				}

<*>^#{W}*echo{W}+\"[^\"\n]+\"{W}*(\177)?\n? {
				  lineno++;
				  *(strrchr(yytext, '\"')+1) = '\0';
				  cp = strchr(yytext, '\"');
				  tagincproc = (char *) mycalloc((size_t)(strlen(cp)+echocmdlen+2), sizeof(char), _taglex_);
				  sprintf(tagincproc, "%s %s", echocmd, cp);
				  tagincfile = (char *) NULL;
				  begintagincfile(INCPROC, tagincfile, tagincproc);
				}

<*>^#{W}*exec{W}+\[[^\n]+\]{W}*(\177)?\n? {
				  if (scanpass == 2) {
					ii = getexecunput(&execunput, yytext, yyleng, FALSE, FALSE, isexpand_macros) - 1;
					while (ii > -1) {
						unput(*(execunput+ii));
						ii--;
					}
					free(execunput);
				  } else {
					lineno++;
				  }
				}

<*>^#{W}*exec{W}+\"\/dev\/null\"{W}+\[[^\n]+\]{W}*(\177)?\n? {
				  if (scanpass == 2) {
					ii = getexecunput(&execunput, yytext, yyleng, FALSE, TRUE, isexpand_macros) - 1;
					while (ii > -1) {
						unput(*(execunput+ii));
						ii--;
					}
					free(execunput);
				  } else {
					lineno++;
				  }
				}

<*>^#{W}*(piktexec|pexec){W}+\[[^\n]+\]{W}*(\177)?\n? {
				  if (scanpass == 2) {
					ii = getexecunput(&execunput, yytext, yyleng, TRUE, FALSE, isexpand_macros) - 1;
					while (ii > -1) {
						unput(*(execunput+ii));
						ii--;
					}
					free(execunput);
				  } else {
					lineno++;
				  }
				}

<*>^#{W}*(piktexec|pexec){W}+\"\/dev\/null\"{W}+\[[^\n]+\]{W}*(\177)?\n? {
				  if (scanpass == 2) {
					ii = getexecunput(&execunput, yytext, yyleng, TRUE, TRUE, isexpand_macros) - 1;
					while (ii > -1) {
						unput(*(execunput+ii));
						ii--;
					}
					free(execunput);
				  } else {
					lineno++;
				  }
				}

<*>^#{W}*(set|setenv){W}+{L}[_0-9a-zA-Z]*{W}*=?{W}*\"[^\"]*\"{W}*(\177)?\n? {
				  lineno++;
				  *strrchr(yytext, '\"') = '\0';
				  cp = strchr(yytext, '\"') + 1;
				  enval = mystrdup(cp, _taglex_);
				  do {
					cp--;
				  } while (! (isalnum(*cp) || (*cp == '_')));
				  *(cp+1) = '\0';
				  do {
					cp--;
				  } while (! isspace(*cp));
				  envar = mystrdup(cp+1, _taglex_);
				  setenv(envar, enval, TRUE);
				  free(envar);
				  free(enval);
				}

<*>^#{W}*(set|setenv){W}+{L}[_0-9a-zA-Z]*{W}*(\177)?\n? {
				  lineno++;
				  cp = yytext+yyleng;
				  do {
					cp--;
				  } while (! (isalnum(*cp) || (*cp == '_')));
				  *(cp+1) = '\0';
				  do {
					cp--;
				  } while (! isspace(*cp));
				  envar = mystrdup(cp+1, _taglex_);
				  setenv(envar, "1", TRUE);
				  free(envar);
				}

<*>^#{W}*(unset|unsetenv){W}+{L}[_0-9a-zA-Z]*{W}*(\177)?\n? {
				  lineno++;
				  cp = yytext+yyleng;
				  do {
					cp--;
				  } while (! (isalnum(*cp) || (*cp == '_')));
				  *(cp+1) = '\0';
				  do {
					cp--;
				  } while (! isspace(*cp));
				  envar = mystrdup(cp+1, _taglex_);
				  unsetenv(envar);
				  free(envar);
				}

<*>^#{W}*if(sys)?{W}+		{ BEGIN(COND);
				  return _IF;
				}
<*>^#{W}*(elif|elsif|elseif)(sys)?{W}+ {
				  BEGIN(COND);
				  return _ELIF;
				}
<*>^#{W}*else(sys)?{W}*\n	{ unput('\n');
				  BEGIN(COND);
				  return _ELSE;
				}
<*>^#{W}*else(sys)?(\177)?\n	{ unput('\n');
				  BEGIN(COND);
				  return _ELSE;
				}
<*>^#{W}*(endif|fi)(sys)?{W}*\n { unput('\n');
				  BEGIN(COND);
				  return _ENDIF;
				}
<*>^#{W}*(endif|fi)(sys)?(\177)?\n {
				  unput('\n');
				  BEGIN(COND);
				  return _ENDIF;
				}

<COND,INC>[^#!&|() \t\n\177]+	{ RETURN(_IDENT); }
<COND,INC>"!"|"&"|"|"|"("|")"	{ return yytext[0]; }
<COND,INC>{W}+			{ ; }
<COND,INC>(\177)?\n		{ lineno++;
				  BEGIN(INITIAL);
				}

<*>^{W}*\n?			{ lineno++;
				  if (indented || ((yyleng > 1) && ((cfgcurr == PRGCFG) || (cfgcurr == FILCFG)))) {
					UPDATE_INCLUDE;
					hasmacro = FALSE;
					taglval.sval = mystrdup("\t\n", _taglex_);
					return _LINE;
				  }
				}

<*>\177\n?			{ lineno++;
				  UPDATE_INCLUDE;
				}

<*>^#{W}*(def|define|undef|undefine|ifdef|ifndef|elifdef|elsifdef|elseifdef|elifndef|elsifndef|elseifndef|elsedef|endifdef|fidef|setdef)[^\n\177]*(\177)?\n? {
				  lineno++;
				  inspectline(yytext, &hasmacro, TRUE);
				  RETURN(_DEF);
				}

<*>^#.?				{ BEGIN(REM);
				  yyless(0);
				}
<*>^ERR.*\n?			{ BEGIN(REM);
				  yyless(3);
				}
<REM>.*\n?			{ lineno++;
				  BEGIN(INITIAL);
				  if (indented) {
					UPDATE_INCLUDE;
					taglval.sval = (char *) mycalloc((size_t)(yyleng+2), sizeof(char), _taglex_);
					sprintf(taglval.sval, "\t%s", yytext);
					return _LINE;
				  } else {
					if ((cp = strchr(yytext,'\n')))
						*cp = '\0';
					tagerror("syntax error");
				  }
				}

<*>^[^#\n\177][^\n\177]*(\177)?\n? {
				  lineno++;
				  UPDATE_INCLUDE;
				  inspectline(yytext, &hasmacro, TRUE);
				  if (indented) {
					taglval.sval = (char *) mycalloc((size_t)(yyleng+2), sizeof(char), _taglex_);
					sprintf(taglval.sval, "\t%s", yytext);
					return _LINE;
				  } else {
					RETURN(_LINE);
				  }
				}

<*>\n				{ /* when pattern matching #indent & #unindent
				     above within #include files, the \n is
				     not captured within yytext; moreover,
				     it somehow "leaks" to stdout, so that,
				     for every #indent & #unindent line,
				     blank lines appear in the piktc command
				     output (before the "processing..." line);
				     we have to suppress the \n here to prevent
				     the glitch; so very, very strange!
				   */
				  *yytext = '\0';
				}

%%

/* tag error handler
 */
void
tagerror (const char *message)
{

	if (! bailout) {
		logreturn(OUTPUT, LOG_ERR, ERRID_TAG_PARSE_ERROR, __FILE__, __LINE__, "tagparse", FALSE, "%s, at or near line %d: %s: %s", cfgfilecurr, lineno, message, yytext ? yytext : nullstring);
	}
	bailout = TRUE;
	if (incnum > 0) {
		while (incnum > 0) {
			endtagincfile();
		}
	} else {
		BEGIN(INITIAL);
	}
	return;
}

/* tag end-of-file handler
 */
int
tagwrap (void)
{

	if (incnum > 0) {
		if (bailout) {
			while (incnum > 0) {
				endtagincfile();
			}
			return 1;
		} else {
			endtagincfile();
			return 0;
		}
	} else {
		return 1;
	}
}

/* push the current input buffer onto the file stack; create and switch to
   a new input buffer; open the new include file and start feeding its
   contents to the new input buffer; if stripping comments from the new
   include file, spawn a child process to pass the include file
   contents to the strip lexer
 */
void
begintagincfile (int inctyp, char *file, char *proc)
{
	static char fnnam[] = "begintagincfile";

	if (    ((cfgcurr == DEFCFG) && (scanpass == 1))
	     || ((cfgcurr  > DEFCFG) && (scanpass == 2))
	   ) {
		check_incfile(inctyp, file);
		if (bailout == TRUE)
			return;
	}
	pushtagincfile(inctype, infile, inproc, inctmpfile, inctmpout, tagin, fd, YY_CURRENT_BUFFER, pid, domacros, indented, lineno);
	inctype = inctyp;
	infile = file;
	inproc = proc;
	if (file && ! strcmp(file, "/dev/null"))
		incdevnull = TRUE;
	else
		incdevnull = FALSE;
	if (UPDATE_INCLUDES) {
		inctmpfile = (char *) mymalloc((strlen(file)+5) * sizeof(char), fnnam);
		sprintf(inctmpfile, "%s.tmp", file);
		if ((inctmpout = fopen(inctmpfile, "w")) == NULL) {
			logreturn(OUTPUT, LOG_ERR, ERRID_FILE_OPEN_ERROR, __FILE__, __LINE__, fnnam, TRUE, "fopen() error for %s", inctmpfile);
			bailout = TRUE;
			return;
		}
		chmod(inctmpfile, (mode_t)(S_IRUSR|S_IWUSR|S_IRGRP));	/* 640 */
	}
	cfgfileprev = cfgfilecurr;
	if (inctype == INCPROC)
		cfgfilecurr = inproc;
	else
		cfgfilecurr = infile;
	if (scanpass != 1) {
		if (verbatim)
			domacros = FALSE;
		else
			domacros = TRUE;
	}
	incnum++;
	if (incnum > MAXIFS) {
		logreturn(OUTPUT, LOG_ERR, ERRID_INCLUDE_FILE_MAX_EXCEEDED, __FILE__, __LINE__, fnnam, FALSE, "maximum number (%d) of include files exceeded", MAXIFS);
		bailout = TRUE;
		return;
	}
	/*
	indented = FALSE;
	 */
	lineno = 1;
	if (verbatim)
		BEGIN(VRB);
	else
		BEGIN(INC);
	if (! verbatim && dostrip) {
		if (pipe(fd) < 0) {
			logreturn(OUTPUT, LOG_ERR, ERRID_PIPE_ERROR, __FILE__, __LINE__, fnnam, TRUE, "pipe() error");
			bailout = TRUE;
			return;
		}
		/* open the ipc channels */
		ipcopen(incnum, __FILE__, __LINE__);
		switch (pid = fork()) {
			case -1:	/* error */
				logreturn(OUTPUT, LOG_ERR, ERRID_FORK_ERROR, __FILE__, __LINE__, fnnam, TRUE, "fork() error");
				bailout = TRUE;
				return;
				break;
			/* pipe from child to parent */
			case 0:		/* child */
				close(fd[0]);
				/* todo: check the return value of instrmopen()
				   and if error, signal parent to bailout
				 */
				stripin = instrmopen(inctype, infile, inproc, incnum, __FILE__, __LINE__);
				stripout = fdopen(fd[1], "w");
				wait_parent(incnum, __FILE__, __LINE__);
				tell_parent(incnum, __FILE__, __LINE__);
				striplex();
				/* todo: check the return value of instrmclose()
				   and if error, signal parent to bailout
				 */
				instrmclose(inctype, stripin, incnum, __FILE__, __LINE__);
				fclose(stripout);
				tell_parent(incnum, __FILE__, __LINE__);
				wait_parent(incnum, __FILE__, __LINE__);
				close(fd[1]);
				exit(0);
				break;
			default:	/* parent */
				close(fd[1]);
				/* the following sleep() is probably not needed
				   but helps prevent hangs on slower, overloaded
				   systems
				sleep(fork_delay);
				 */
				tagin = fdopen(fd[0], "r");
				tell_child(incnum, __FILE__, __LINE__);
				wait_child(incnum, __FILE__, __LINE__);
				break;
		}
	} else {
		pid = 0;
		fd[0] = -1;
		fd[1] = -1;
		if (! (tagin = instrmopen(inctype, infile, inproc, incnum, __FILE__, __LINE__))) {
			bailout = TRUE;
			return;
		}
	}
	yy_switch_to_buffer(yy_create_buffer(tagin, MAXBUF));
	return;
}

/* at the end of the current include file, close the file, delete its
   input buffer, pop the previous include file off the file stack, and
   switch to the previous input buffer
 */
void
endtagincfile (void)
{
	static char fnnam[] = "endtagincfile";
	YY_BUFFER_STATE bs;

	if (! verbatim && dostrip) {
		wait_child(incnum, __FILE__, __LINE__);
		tell_child(incnum, __FILE__, __LINE__);
		close(fd[0]);
		/* close the ipc channels */
		ipcclose(incnum, __FILE__, __LINE__);
	} else {
		if (instrmclose(inctype, tagin, incnum, __FILE__, __LINE__) == ERROR)
			bailout = TRUE;
	}
	if (UPDATE_INCLUDES) {
		if (fclose(inctmpout)) {
			logreturn(OUTPUT, LOG_ERR, ERRID_FILE_CLOSE_ERROR, __FILE__, __LINE__, fnnam, TRUE, "fclose() error for %s", inctmpfile);
			bailout = TRUE;
		}
		if (! bailout) {
			if (rename(inctmpfile, infile)) {
				logreturn(OUTPUT, LOG_ERR, ERRID_FILE_RENAME_ERROR, __FILE__, __LINE__, fnnam, TRUE, "rename() error for %s to %s", inctmpfile, infile);
				bailout = TRUE;
			}
		} else {
			unlink(inctmpfile);
		}
	}
	verbatim = FALSE;
	incnum--;
	if (incnum < 0) {
		logreturn(OUTPUT, LOG_ERR, ERRID_INCNUM_LESS_THAN_ZERO, __FILE__, __LINE__, fnnam, FALSE, "incnum < 0");
		bailout = TRUE;
		return;
	}
	yy_delete_buffer(YY_CURRENT_BUFFER);
	poptagincfile(&inctype, &infile, &inproc, &inctmpfile, &inctmpout, &tagin, fd, &bs, &pid, &domacros, &indented, &lineno);
	cfgfileprev = cfgfilecurr;
	if (inctype == INCPROC)
		cfgfilecurr = inproc;
	else
		cfgfilecurr = infile;
	yy_switch_to_buffer(bs);
	if (incnum == 0)
		BEGIN(INITIAL);
	else
		BEGIN(INC);
	return;
}

/* push current input buffer info onto the file stack
 */
void
pushtagincfile (int inctyp, char *file, char *proc, char *tmpfile, FILE *tmpout, FILE *yyin, int *fd, YY_BUFFER_STATE bs, pid_t pid, int domcr, int indnt, int lno)
{

	tagincfiles[incnum].inctyp = inctyp;
	tagincfiles[incnum].file = file;
	tagincfiles[incnum].proc = proc;
	tagincfiles[incnum].tmpfile = tmpfile;
	tagincfiles[incnum].tmpout = tmpout;
	tagincfiles[incnum].yyin = yyin;
	tagincfiles[incnum].fd[0] = *fd;
	tagincfiles[incnum].fd[1] = *(fd+1);
	tagincfiles[incnum].bs = bs;
	tagincfiles[incnum].pid = pid;
	tagincfiles[incnum].domcr = domcr;
	tagincfiles[incnum].indnt = indnt;
	tagincfiles[incnum].lineno = lno;
	return;
}

/* pop input buffer info off the file stack
 */
void
poptagincfile (int *inctyp, char **file, char **proc, char **tmpfile, FILE **tmpout, FILE **yyin, int *fdp, YY_BUFFER_STATE *bsp, pid_t *pidp, int *dmp, int *indntp, int *lnp)
{

	*inctyp = tagincfiles[incnum].inctyp;
	*file = tagincfiles[incnum].file;
	*proc = tagincfiles[incnum].proc;
	*tmpfile = tagincfiles[incnum].tmpfile;
	*tmpout = tagincfiles[incnum].tmpout;
	*yyin = tagincfiles[incnum].yyin;
	*fdp = tagincfiles[incnum].fd[0];
	*(fdp+1) = tagincfiles[incnum].fd[1];
	*bsp = tagincfiles[incnum].bs;
	*pidp = tagincfiles[incnum].pid;
	*dmp = tagincfiles[incnum].domcr;
	*indntp = tagincfiles[incnum].indnt;
	*lnp = tagincfiles[incnum].lineno;
	return;
}

/* determine if the given line might have at least one macro embedded within;
   we don't check for escapes here, so we might get false positives; still,
   we have to set *hasmcr to TRUE in such cases to force the proper \=
   handling later on in expand();

   also, strip out any '\177' char if so indicated
 */
void inspectline (char *s, int *hasmcr, int strip)
{
	register int i;
	register char c1, c2;

	*hasmcr = FALSE;
	i = 0;
	while ((c1 = *(s+i))) {
		if ((c1 == '\177') && strip) {
			/* if strip, dispose of any '\177' */
			*(s+i) = *(s+i+1);
			*(s+i+1) = '\0';
			break;
		} else if (    (c1 == '=')
		            && (c2 = *(s+i+1))
			    && (   isalpha((int) c2)
			         || (c2 == '_')
			       )
			   ) {
				*hasmcr = TRUE;
		}
		i++;
	}
	return;
}
