%{

/*
   cmd.l -- PIKT pikt command-line lexer

   $Header: /var/cvs/PIKT/pikt/src/piktd/cmd.l,v 1.38 2007/01/07 02:33:20 berto Exp $

   This file is part of PIKT ("Problem Informant/Killer Tool").
   Copyright (C) 1998-2006 by Robert Osterlund.

   PIKT is a trademark of the University of Chicago.  All Rights Reserved.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   PIKT is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
   USA.

   Send problem reports, bouquets and brickbats to:  pikt-bugs@pikt.org
 */

#include "../lib/common.h"
#include "piktd.h"
#include "cmd.tab.h"

char _cmdlex_[] = "cmdlex";

#undef YY_INPUT
#define YY_INPUT(buf, result, max_size)	{ \
		signed char c = cmdinput(); \
		result = (c == EOF) ? YY_NULL : (buf[0] = c, 1); \
	}

char errormsg[MAXLINE+1];

int isqtstr;		/* does this command-line argument
			   appear to be a quoted string?
			 */

int iscommand = FALSE;

void cmderror(const char *);
signed int cmdinput(void);

%}

W		[ \t]
nW		[^ \t]
D		[0-9]
L		[A-Za-z]
DL		[0-9A-Za-z]

%x cS mS pS sS

%option noyywrap

%%

<INITIAL,mS,pS>"-"		{ BEGIN(mS);
				  return _MINUS;
				}
<INITIAL,mS,pS>"+"		{ BEGIN(pS);
				  return _PLUS;
				}

<mS>h				{ ishelp = TRUE;
				  return _OPT;
				}
<mS>G				{ isdebug = TRUE;
				  return _OPT;
				}
<mS>T				{ istest = TRUE;
				  /* isdebug = TRUE; */
				  return _OPT;
				}
<mS>V				{ isshow_version = TRUE;
				  return _OPT;
				}
<mS>z				{ ismakecheck = TRUE;
				  isdebug = TRUE;
				  return _OPT;
				}

<mS>c				{ ischeck = TRUE;
				  return _OPT;
				}
<mS>C				{ if (! hostmatch(hostname, master)) {
					logexit(PIKTLOG, LOG_ERR, ERRID_ATTEMPTED_DOUBLECHECK_ON_NON_MASTER, __FILE__, __LINE__, _cmdlex_, FALSE, "attempted doublecheck on non-master system");
				  }
				  interactive = TRUE;
				  isdoublecheck = TRUE;
				  ismakecheck = TRUE;
				  return _OPT;
				}

<pS>M				{ return _MAILCMD; }
<pS>L				{ return _LPCMD; }

<pS>A				{ return _ALERT; }

<pS>S				{ BEGIN(sS);
				  return _SCRIPT;
				}

<pS>C				{ BEGIN(cS);
				  iscommand = TRUE;
				  return _COMMAND;
				}

<*>{W}				{ ; }

<INITIAL,pS,sS>\"[^\"]*\"	{ yytext[yyleng-1] = '\0';
				  cmdlval.sval = mystrdup(yytext+1, _cmdlex_);
				  return _ITEM;
				}

<cS>\'[^\']*\'			{ yytext[yyleng-1] = '\0';
				  cmdlval.sval = mystrdup(yytext+1, _cmdlex_);
				  return _ITEM;
				}

<sS>{nW}+			{ cmdlval.sval = mystrdup(yytext, _cmdlex_);
				  return _ITEM;
				}

<INITIAL,pS>[^-+ \t\"][^ \t\"]*	{ cmdlval.sval = mystrdup(yytext, _cmdlex_);
				  return _ITEM;
				}

<INITIAL>"-"[^hGzTVcC]{nW}*	{ cmdlval.sval = mystrdup(yytext, _cmdlex_);
				  return _ITEM;
				}

<INITIAL>"+"[^MLASC]{nW}*	{ cmdlval.sval = mystrdup(yytext, _cmdlex_);
				  return _ITEM;
				}

<mS,pS>.			{ sprintf(errormsg, "illegal option: %c", yytext[0]);
				  cmderror(errormsg);
				  ishelp = TRUE;
				  return _OPT;
				 }

%%

/* cmd error handler
 */
void
cmderror (const char *message)
{

	logreturn(OUTPUT, LOG_ERR, ERRID_CMD_PARSE_ERROR, __FILE__, __LINE__, _cmdlex_, FALSE, "%s: %s", message, yytext ? yytext : nullstring);
	bailout = TRUE;
	return;
}

/* feed command-line arguments to the lexer input stream
 */
signed int
cmdinput (void)
{
	register char c;

	if (argvcurr > argvlast)
		return(EOF);
	if ((stroffset == 0) && ! isqtstr) {
		if (strchr(*argvcurr, ' ') || strchr(*argvcurr, '\t') || ! strcmp(*argvcurr, "quit")) {
			isqtstr = TRUE;
			return(iscommand ? '\'' : '"');
		}
	}
	if ((c = (*argvcurr)[stroffset]) != '\0') {
		stroffset++;
		return(c);
	}
	if (isqtstr) {
		isqtstr = FALSE;
		return(iscommand ? '\'' : '"');			
	}
	/* end of argument, move to the next */
	argvcurr++;
	stroffset = 0;
	return(' ');
}
