%{

/*
   hst.l -- PIKT history file lexer

   $Header: /var/cvs/PIKT/pikt/src/piktd/hst.l,v 1.30 2006/11/04 09:59:36 berto Exp $

   This file is part of PIKT ("Problem Informant/Killer Tool").
   Copyright (C) 1998-2006 by Robert Osterlund.

   PIKT is a trademark of the University of Chicago.  All Rights Reserved.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   PIKT is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
   USA.

   Send problem reports, bouquets and brickbats to:  pikt-bugs@pikt.org

   ----------------------------------------------------------------------------

   on our test irix system, flex failed to match regexps of the form
   '{D}{x,y}', which led to parser failures; hence substitutions of the
   form '{D}+' below

 */

#include "../lib/common.h"
#include "piktd.h"
#include "hst.tab.h"

char _hstlex_[] = "hstlex";

int lineno;
int ii;

void hsterror(const char *);

%}

W		[ \t]
D		[0-9]
L		[A-Za-z]
DL		[0-9A-Za-z]
nB		[^\[\]]

%x RS KS LS VS SS

%option noyywrap

%%

<*>^A			{ BEGIN(INITIAL);
			  return _ALM;
			}

<*>^L			{ BEGIN(LS);
			  return _LOG;
			}

<*>^R			{ ii = 0;
			  BEGIN(RS);
			  return _VAR;
			}

<*>^K			{ BEGIN(KS);
			  return _KEY;
			}

<*>^V			{ BEGIN(VS);
			  return _VAL;
			}

<LS>{W}{D}+		{ hstlval.lval = atol(yytext+1);
			  return _LNG;
			}
<LS>{W}[^0-9].*$	{ hstlval.sval = mystrdup(yytext+1, _hstlex_);
			  BEGIN(INITIAL);
			  return _STR;
			}

<RS>{W}{D}+		{ hstlval.ival = atoi(yytext+1);
			  ii++;
			  if (ii >= 4)
				BEGIN(SS);
			  return _INT;
			}

<KS>{W}{D}+		{ hstlval.ival = atoi(yytext+1);
			  return _INT;
			}
<KS>{W}"["{nB}*"]"	{ hstlval.sval = mystrdup(yytext+2, _hstlex_);
			  *(hstlval.sval+yyleng-3) = '\0';
			  BEGIN(INITIAL);
			  return _KST;
			}

<VS>{W}{D}+		{ hstlval.ival = atoi(yytext+1);
			  return _INT;
			}
<VS>{W}"["{nB}*"]"	{ hstlval.sval = mystrdup(yytext+2, _hstlex_);
			  *(hstlval.sval+yyleng-3) = '\0';
			  BEGIN(SS);
			  return _KST;
			}

<INITIAL,SS>{W}.*$	{ hstlval.sval = mystrdup(yytext+1, _hstlex_);
			  BEGIN(INITIAL);
			  return _STR;
			}

<*>\n			{ lineno++; }
<*>.			{ return yytext[0]; }

%%

/* hst error handler
 */
void
hsterror (const char *message)
{
	static int l = 0;

	if (l != lineno)
		l = lineno;
	/* if lineno > 1, have successfully parsed the A line, hence have
	   set alarmcurr to match the current alarm history, and have set
	   alarmcurr->hsterr = 0; if lineno = 1, alarmcurr is undefined
	 */
	if (lineno > 1)
		HSTERR = 1;
	logreturn(PIKTLOG, LOG_ERR, ERRID_HST_PARSE_ERROR, __FILE__, __LINE__, _hstlex_, FALSE, "in .hst file, at or near line %d: %s: @%s@", lineno, message, yytext ? yytext : nullstring);
	return;
}
